﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;

namespace $rootnamespace$.HojasCalculo.Excel
{
    internal static class ComposicionExcel
    {
        public static void Componer(ExcelApplication excel, Worksheet hoja, IEnumerable<ICelda> celdas)
        {
            var visitanteTamano = new VisitanteTamano();
            foreach (var celda in celdas)
                celda.Aceptar(visitanteTamano);

            var maxFilas = visitanteTamano.MaxFila;
            var maxColumnas = visitanteTamano.MaxColumna;

            using (var rangoA1 = hoja.Cells[1, 1]) {
                using (var rango = rangoA1.Resize(maxFilas, maxColumnas)) {
                    using (var columnas = rango.EntireColumn) {
                        columnas.ColumnWidth = 12;
                    }
                    using (var fuente = rango.Font) {
                        fuente.Name = "Times New Roman";
                    }
                }
            }

            var visitante = new Visitante(hoja);
            foreach (var celda in celdas)
                celda.Aceptar(visitante);

            var visitanteBordes = new VisitanteBordes(maxFilas, maxColumnas);
            foreach (var celda in celdas) {
                celda.Aceptar(visitanteBordes);
            }

            DibujarBordes(hoja, visitanteBordes.Bordes);

            ConfigurarPagina(hoja, maxFilas, maxColumnas);
        }

        private static void ConfigurarPagina(Worksheet hoja, int maxFilas, int maxColumnas)
        {
            using (var configuracionPagina = hoja.PageSetup) {
                configuracionPagina.FitToPagesTall = false;
                configuracionPagina.FitToPagesWide = 1;

                configuracionPagina.Zoom = false;

                if (maxFilas != 1 || maxColumnas != 1)
                    configuracionPagina.PrintArea = string.Format("{0}:{1}", hoja.Cells[1, 1].Address(true, true), hoja.Cells[maxFilas, maxColumnas].Address(true, true));

                configuracionPagina.LeftMargin = 0;
                configuracionPagina.TopMargin = 0;
                configuracionPagina.RightMargin = 0;
                configuracionPagina.BottomMargin = 0;
                configuracionPagina.HeaderMargin = 0;
                configuracionPagina.FooterMargin = 0;
            }
        }

        private static void DibujarBordes(Worksheet hoja, BordesCelda[,] bordes)
        {
            for (var i = 0; i < bordes.GetLength(0); i++) {
                for (var j = 0; j < bordes.GetLength(1); j++) {
                    var borde = bordes[i, j];
                    if (borde.Superior.Grosor != GrosorBorde.Ninguno || borde.Izquierdo.Grosor != GrosorBorde.Ninguno || borde.Inferior.Grosor != GrosorBorde.Ninguno || borde.Derecho.Grosor != GrosorBorde.Ninguno) {
                        using (var celda = hoja.Cells[i + 1, j + 1]) {
                            if (borde.Superior.Grosor != GrosorBorde.Ninguno) {
                                var grosor = borde.Superior.Grosor == GrosorBorde.Fino ? XlBorderWeight.xlThin : (borde.Superior.Grosor == GrosorBorde.Medio ? XlBorderWeight.xlMedium : XlBorderWeight.xlThick);

                                using (var superior = celda.Borders[XlBordersIndex.xlEdgeTop]) {
                                    superior.Weight = grosor;
                                    superior.LineStyle = XlLineStyle.xlContinuous;
                                    superior.ColorIndex = borde.Superior.Color;
                                }
                            }
                            if (borde.Izquierdo.Grosor != GrosorBorde.Ninguno) {
                                var grosor = borde.Izquierdo.Grosor == GrosorBorde.Fino ? XlBorderWeight.xlThin : (borde.Izquierdo.Grosor == GrosorBorde.Medio ? XlBorderWeight.xlMedium : XlBorderWeight.xlThick);

                                using (var izquierdo = celda.Borders[XlBordersIndex.xlEdgeLeft]) {
                                    izquierdo.Weight = grosor;
                                    izquierdo.LineStyle = XlLineStyle.xlContinuous;
                                    izquierdo.ColorIndex = borde.Izquierdo.Color;
                                }
                            }
                            if (borde.Inferior.Grosor != GrosorBorde.Ninguno) {
                                var grosor = borde.Inferior.Grosor == GrosorBorde.Fino ? XlBorderWeight.xlThin : (borde.Inferior.Grosor == GrosorBorde.Medio ? XlBorderWeight.xlMedium : XlBorderWeight.xlThick);

                                using (var inferior = celda.Borders[XlBordersIndex.xlEdgeBottom]) {
                                    inferior.Weight = grosor;
                                    inferior.LineStyle = XlLineStyle.xlContinuous;
                                    inferior.ColorIndex = borde.Inferior.Color;
                                }
                            }
                            if (borde.Derecho.Grosor != GrosorBorde.Ninguno) {
                                var grosor = borde.Derecho.Grosor == GrosorBorde.Fino ? XlBorderWeight.xlThin : (borde.Derecho.Grosor == GrosorBorde.Medio ? XlBorderWeight.xlMedium : XlBorderWeight.xlThick);

                                using (var derecho = celda.Borders[XlBordersIndex.xlEdgeRight]) {
                                    derecho.Weight = grosor;
                                    derecho.LineStyle = XlLineStyle.xlContinuous;
                                    derecho.ColorIndex = borde.Derecho.Color;
                                }
                            }
                        }
                    }
                }
            }
        }

        private class VisitanteTamano : IVisitante
        {
            public int MaxFila { get; private set; }
            public int MaxColumna { get; private set; }

            public VisitanteTamano()
            {
                this.MaxFila = 1;
                this.MaxColumna = 1;
            }

            public void Visitar(Celda celda)
            {
                MaxFila = Math.Max(MaxFila, celda.Dir1.Fila);
                MaxColumna = Math.Max(MaxColumna, celda.Dir1.Columna);

                MaxFila = Math.Max(MaxFila, celda.Dir2.Fila);
                MaxColumna = Math.Max(MaxColumna, celda.Dir2.Columna);
            }

            public void Visitar(Valor valor)
            {
                valor.Celda.Aceptar(this);
            }

            public void Visitar(Formato formato)
            {
                formato.Celda.Aceptar(this);
            }

            public void Visitar(Negrita negrita)
            {
                negrita.Celda.Aceptar(this);
            }

            public void Visitar(Subrayado subrayado)
            {
                subrayado.Celda.Aceptar(this);
            }

            public void Visitar(Cursiva cursiva)
            {
                cursiva.Celda.Aceptar(this);
            }

            public void Visitar(Fuente fuente)
            {
                fuente.Celda.Aceptar(this);
            }

            public void Visitar(Tamano tamano)
            {
                tamano.Celda.Aceptar(this);
            }

            public void Visitar(ColorFondo fondo)
            {
                fondo.Celda.Aceptar(this);
            }

            public void Visitar(ColorFuente colorFuente)
            {
                colorFuente.Celda.Aceptar(this);
            }

            public void Visitar(Centrado centrado)
            {
                centrado.Celda.Aceptar(this);
            }

            public void Visitar(Derecha derecha)
            {
                derecha.Celda.Aceptar(this);
            }

            public void Visitar(VCentrado centrado)
            {
                centrado.Celda.Aceptar(this);
            }

            public void Visitar(AjustarTexto ajustado)
            {
                ajustado.Celda.Aceptar(this);
            }

            public void Visitar(Orientacion orientacion)
            {
                orientacion.Celda.Aceptar(this);
            }

            public void Visitar(Bordeado bordeado)
            {
                bordeado.Celda.Aceptar(this);
            }

            public void Visitar(ColorBorde colorBorde)
            {
                colorBorde.Celda.Aceptar(this);
            }

            public void Visitar(Combinada combinada)
            {
                combinada.Celda.Aceptar(this);
            }

            public void Visitar(Imagen imagen)
            {
                imagen.Celda.Aceptar(this);
            }

            public void Visitar(AltoFila alto)
            {
                alto.Celda.Aceptar(this);
            }

            public void Visitar(AnchoColumna ancho)
            {
                ancho.Celda.Aceptar(this);
            }
        }

        private class Visitante : IVisitante
        {
            private Worksheet hoja;
            private Range rango;

            public Visitante(Worksheet hoja)
            {
                this.hoja = hoja;
            }

            public void Visitar(Celda celda)
            {
                rango = hoja.Range(celda.Dir1, celda.Dir2);
            }

            public void Visitar(Valor valor)
            {
                valor.Celda.Aceptar(this);
                var s = valor.Objeto as string;
                if (s != null) {
                    if (s != "'")
                        rango.Value = s;
                }
                else
                    rango.Value = valor.Objeto;
            }

            public void Visitar(Formato formato)
            {
                formato.Celda.Aceptar(this);
                rango.NumberFormat = formato.Patron;
            }

            public void Visitar(Negrita negrita)
            {
                negrita.Celda.Aceptar(this);
                rango.Font.Bold = true;
            }

            public void Visitar(Subrayado subrayado)
            {
                subrayado.Celda.Aceptar(this);
                rango.Font.Underline = true;
            }

            public void Visitar(Cursiva cursiva)
            {
                cursiva.Celda.Aceptar(this);
                rango.Font.Italic = true;
            }

            public void Visitar(Fuente fuente)
            {
                fuente.Celda.Aceptar(this);
                rango.Font.Name = fuente.Tipo;
            }

            public void Visitar(Tamano tamano)
            {
                tamano.Celda.Aceptar(this);
                rango.Font.Size = tamano.Puntos;
            }

            public void Visitar(ColorFondo fondo)
            {
                fondo.Celda.Aceptar(this);
                rango.Interior.ColorIndex = fondo.Color;
            }

            public void Visitar(ColorFuente colorFuente)
            {
                colorFuente.Celda.Aceptar(this);
                rango.Font.ColorIndex = colorFuente.Color;
            }

            public void Visitar(Centrado centrado)
            {
                centrado.Celda.Aceptar(this);
                rango.HorizontalAlignment = HorizontalAlignment.xlCenter;
            }

            public void Visitar(Derecha derecha)
            {
                derecha.Celda.Aceptar(this);
                rango.HorizontalAlignment = HorizontalAlignment.xlRight;
            }

            public void Visitar(VCentrado centrado)
            {
                centrado.Celda.Aceptar(this);
                rango.VerticalAlignment = VerticalAlignment.xlCenter;
            }

            public void Visitar(AjustarTexto ajustado)
            {
                ajustado.Celda.Aceptar(this);
                rango.WrapText = true;
            }

            public void Visitar(Orientacion orientacion)
            {
                orientacion.Celda.Aceptar(this);
                rango.Orientation = orientacion.Angulo;
            }

            public void Visitar(Bordeado bordeado)
            {
                bordeado.Celda.Aceptar(this);
            }

            public void Visitar(ColorBorde colorBorde)
            {
                colorBorde.Celda.Aceptar(this);
            }

            public void Visitar(Combinada combinada)
            {
                combinada.Celda.Aceptar(this);
                rango.Merge();
                rango.VerticalAlignment = VerticalAlignment.xlTop;
            }

            public void Visitar(Imagen imagen)
            {
                imagen.Celda.Aceptar(this);

                var dirTemp = Path.GetTempPath();
                var ficheroTemp = Path.ChangeExtension(Path.Combine(dirTemp, Path.GetFileName(imagen.Ruta)), "jpg");

                int ancho, alto;
                try {
                    using (var i = Image.FromFile(imagen.Ruta)) {
                        using (var b = ReducirImagen(i, (int)rango.Width - 4, (int)rango.Height - 4)) {
                            ancho = b.Width;
                            alto = b.Height;
                            b.Save(ficheroTemp, ImageFormat.Jpeg);
                        }
                    }
                }
                catch (FileNotFoundException) {
                    return;
                }
                hoja.Shapes.AddPicture(ficheroTemp, MsoTriState.msoTrue, MsoTriState.msoTrue, (float)rango.Left + 2f, (float)rango.Top + 2f, ancho, alto);
            }

            private static Bitmap ReducirImagen(Image imagen, int anchoMaximo, int altoMaximo)
            {
                var ratioX = anchoMaximo / (double)imagen.Width;
                var ratioY = altoMaximo / (double)imagen.Height;
                var ratio = Math.Min(ratioX, ratioY);
                ratio = Math.Min(ratio, 1);

                var ancho = (int)(imagen.Width * ratio);
                var alto = (int)(imagen.Height * ratio);

                var xPosition = (int)((anchoMaximo - ancho) / 2);
                var yPosition = (int)((altoMaximo - alto) / 2);

                var imagenReducida = new Bitmap(ancho, alto);
                using (var graphics = Graphics.FromImage(imagenReducida)) {
                    graphics.DrawImage(imagen, 0, 0, ancho, alto);
                }

                return imagenReducida;
            }

            public void Visitar(AltoFila alto)
            {
                alto.Celda.Aceptar(this);
                rango.RowHeight = alto.Puntos;
            }

            public void Visitar(AnchoColumna ancho)
            {
                ancho.Celda.Aceptar(this);
                rango.ColumnWidth = ancho.Caracteres;
            }
        }

        private struct BordeCelda
        {
            public GrosorBorde Grosor { get; set; }
            public int Color { get; set; }
        }

        private struct BordesCelda
        {
            public BordeCelda Superior { get; set; }
            public BordeCelda Izquierdo { get; set; }
            public BordeCelda Inferior { get; set; }
            public BordeCelda Derecho { get; set; }
        }

        private class VisitanteBordes : IVisitante
        {
            private int fila1, columna1;
            private int fila2, columna2;
            public BordesCelda[,] Bordes { get; private set; }

            public VisitanteBordes(int maxFila, int maxColumna)
            {
                Bordes = new BordesCelda[maxFila, maxColumna];
            }

            public void Visitar(Celda celda)
            {
                fila1 = celda.Dir1.Fila;
                columna1 = celda.Dir1.Columna;
                fila2 = celda.Dir2.Fila;
                columna2 = celda.Dir2.Columna;
            }

            public void Visitar(Valor valor)
            {
                valor.Celda.Aceptar(this);
            }

            public void Visitar(Formato formato)
            {
                formato.Celda.Aceptar(this);
            }

            public void Visitar(Negrita negrita)
            {
                negrita.Celda.Aceptar(this);
            }

            public void Visitar(Subrayado subrayado)
            {
                subrayado.Celda.Aceptar(this);
            }

            public void Visitar(Cursiva cursiva)
            {
                cursiva.Celda.Aceptar(this);
            }

            public void Visitar(Tamano tamano)
            {
                tamano.Celda.Aceptar(this);
            }

            public void Visitar(Fuente fuente)
            {
                fuente.Celda.Aceptar(this);
            }

            public void Visitar(ColorFondo fondo)
            {
                fondo.Celda.Aceptar(this);
            }

            public void Visitar(ColorFuente colorFuente)
            {
                colorFuente.Celda.Aceptar(this);
            }

            public void Visitar(Centrado centrado)
            {
                centrado.Celda.Aceptar(this);
            }

            public void Visitar(Derecha derecha)
            {
                derecha.Celda.Aceptar(this);
            }

            public void Visitar(VCentrado centrado)
            {
                centrado.Celda.Aceptar(this);
            }

            public void Visitar(AjustarTexto ajustado)
            {
                ajustado.Celda.Aceptar(this);
            }

            public void Visitar(Orientacion orientacion)
            {
                orientacion.Celda.Aceptar(this);
            }

            public void Visitar(Bordeado bordeado)
            {
                bordeado.Celda.Aceptar(this);

                for (var i = fila1; i <= fila2; i++) {
                    if (Bordes[i - 1, columna1 - 1].Izquierdo.Grosor < bordeado.Grosor && (bordeado.Borde & Borde.Izquierdo) == Borde.Izquierdo)
                        Bordes[i - 1, columna1 - 1].Izquierdo = new BordeCelda() { Grosor = bordeado.Grosor, Color = Bordes[i - 1, columna1 - 1].Izquierdo.Color };
                    if (Bordes[i - 1, columna2 - 1].Derecho.Grosor < bordeado.Grosor && (bordeado.Borde & Borde.Derecho) == Borde.Derecho)
                        Bordes[i - 1, columna2 - 1].Derecho = new BordeCelda() { Grosor = bordeado.Grosor, Color = Bordes[i - 1, columna1 - 1].Derecho.Color };
                }
                for (var i = columna1; i <= columna2; i++) {
                    if (Bordes[fila1 - 1, i - 1].Superior.Grosor < bordeado.Grosor && (bordeado.Borde & Borde.Superior) == Borde.Superior)
                        Bordes[fila1 - 1, i - 1].Superior = new BordeCelda() { Grosor = bordeado.Grosor, Color = Bordes[fila1 - 1, i - 1].Superior.Color };
                    if (Bordes[fila2 - 1, i - 1].Inferior.Grosor < bordeado.Grosor && (bordeado.Borde & Borde.Inferior) == Borde.Inferior)
                        Bordes[fila2 - 1, i - 1].Inferior = new BordeCelda() { Grosor = bordeado.Grosor, Color = Bordes[fila1 - 1, i - 1].Inferior.Color };
                }
            }

            public void Visitar(ColorBorde colorBorde)
            {
                colorBorde.Celda.Aceptar(this);

                for (var i = fila1; i <= fila2; i++) {
                    Bordes[i - 1, columna1 - 1].Izquierdo = new BordeCelda() { Grosor = Bordes[i - 1, columna1 - 1].Izquierdo.Grosor, Color = colorBorde.Color };
                    Bordes[i - 1, columna2 - 1].Derecho = new BordeCelda() { Grosor = Bordes[i - 1, columna2 - 1].Derecho.Grosor, Color = colorBorde.Color };
                }
                for (var i = columna1; i <= columna2; i++) {
                    Bordes[fila1 - 1, i - 1].Superior = new BordeCelda() { Grosor = Bordes[fila1 - 1, i - 1].Superior.Grosor, Color = colorBorde.Color };
                    Bordes[fila2 - 1, i - 1].Inferior = new BordeCelda() { Grosor = Bordes[fila2 - 1, i - 1].Inferior.Grosor, Color = colorBorde.Color };
                }
            }

            public void Visitar(Combinada combinada)
            {
                combinada.Celda.Aceptar(this);
            }

            public void Visitar(Imagen imagen)
            {
                imagen.Celda.Aceptar(this);
            }

            public void Visitar(AltoFila alto)
            {
                alto.Celda.Aceptar(this);
            }

            public void Visitar(AnchoColumna ancho)
            {
                ancho.Celda.Aceptar(this);
            }
        }
    }
}
